from importlib.metadata import PackageNotFoundError, version

from lexorank.lexorank import (
    BUCKET_SEPARATOR,
    DECIMAL_POINT,
    WHOLE_NUMBER_SIZE,
    Bucket,
    LexoRank,
    between,
    middle,
    parse,
)

try:
    __version__ = version("lexorank-py")

except PackageNotFoundError:
    pass



from typing_extensions import Self

from lexorank import integer
from lexorank.base import Base, Base36
from lexorank.integer import Integer, Sign

DECIMAL_POINT = ":"


class Decimal:
    def __init__(
        self, significant_figures: Integer, exponent: int, *, decimal_point: str = DECIMAL_POINT
    ) -> None:
        self._integer, self._exponent = self._rstrip(significant_figures, exponent)
        self._decimal_point = decimal_point

    @property
    def base(self) -> Base:
        return self._integer.base

    @property
    def decimal_point(self) -> str:
        return self._decimal_point

    def __neg__(self) -> Self:
        return self.__class__(-self._integer, self._exponent)

    def __abs__(self) -> Self:
        return self.__class__(abs(self._integer), self._exponent)

    def __add__(self, other: object) -> Self:
        other = self._type_guard(other)

        if self._exponent == other._exponent:
            return self.__class__(self._integer + other._integer, self._exponent)

        lv: Decimal = self
        rv = other
        if self._exponent < other._exponent:
            lv, rv = rv, lv

        diff = lv._exponent - rv._exponent
        lv_integer = lv._integer << diff
        lv_exponent = lv._exponent - diff

        return self.__class__(lv_integer + rv._integer, lv_exponent)

    def __sub__(self, other: object) -> Self:
        other = self._type_guard(other)

        return self + (-other)

    def __mul__(self, other: object) -> Self:
        other = self._type_guard(other)

        return self.__class__(self._integer * other._integer, self._exponent + other._exponent)

    def __eq__(self, other: object) -> bool:
        other = self._type_guard(other)

        return self._integer == other._integer and self._exponent == other._exponent

    def __gt__(self, other: object) -> bool:
        other = self._type_guard(other)

        if self._exponent == other._exponent:
            return self._integer > other._integer
        if self._exponent > other._exponent:
            return self._integer << (self._exponent - other._exponent) > other._integer
        return self._integer > other._integer << (other._exponent - self._exponent)

    def __ge__(self, other: object):
        return self.__gt__(other) or self.__eq__(other)

    def __lt__(self, other: object):
        return not self.__ge__(other)

    def __le__(self, other: object):
        return not self.__gt__(other)

    def whole_number(self) -> Integer:
        if self._exponent < 0:
            return self._integer >> abs(self._exponent)
        return self._integer << self._exponent

    def decimal(self) -> "Decimal":
        return self - self.whole_number()

    def __str__(self) -> str:
        sign = "" if self._integer.sign == Sign.POSITIVE else "-"
        intger_str = str(abs(self._integer))
        if self._exponent < 0:
            if len(intger_str) <= abs(self._exponent):
                intger_str = "0" * (abs(self._exponent) - len(intger_str) + 1) + intger_str
            return (
                sign
                + intger_str[: self._exponent]
                + self._decimal_point
                + intger_str[self._exponent :]
            )
        if self._exponent > 0:
            return sign + intger_str + "0" * self._exponent + self._decimal_point
        return sign + intger_str + self._decimal_point

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} value={self}, base={self.base.base()}>"

    def __int__(self) -> int:
        return int(self.whole_number())

    def __float__(self) -> float:
        return float(int(self._integer) * (self.base.base() ** self._exponent))

    def __len__(self) -> int:
        return len(self._integer)

    @staticmethod
    def _rstrip(value: Integer, exponent: int) -> tuple[Integer, int]:
        index = 0
        for digit in reversed(value.digits):
            if digit != 0:
                break
            index += 1

        return value >> index, exponent + index

    def _type_guard(self, other: object) -> "Decimal":
        if isinstance(other, (int, str, float, Integer)):
            return parse(other, self.base)
        if isinstance(other, Decimal):
            return other
        raise ValueError(f"unsupported operand type(s) {type(other)}")


def parse(
    value: str | int | float | Integer, base: Base = Base36, *, decimal_point: str = DECIMAL_POINT
) -> Decimal:
    if isinstance(value, int):
        return Decimal(integer.parse(value, base), 0, decimal_point=decimal_point)

    if isinstance(value, float):
        # WARN: This is not accurate.
        whole = integer.parse(int(value), base)
        if len(whole) >= 10 or value == 0:
            return Decimal(whole, 0, decimal_point=decimal_point)
        if abs(value) >= 1:
            exponent = len(whole) - 10
        else:
            # TODO: more accurate
            exponent = -9
            v = abs(value)
            while True:
                v *= base.base()
                exponent -= 1
                if v >= 1:
                    break
                if exponent <= -150:
                    break
        return Decimal(
            integer.parse(int(value * (base.base() ** -exponent)), base),
            exponent,
            decimal_point=decimal_point,
        )

    if isinstance(value, Integer):
        return Decimal(value, 0, decimal_point=decimal_point)

    try:
        index = value.index(decimal_point)
    except ValueError:
        return Decimal(integer.parse(value, base), 0, decimal_point=decimal_point)

    integer_str = value[:index] + value[index + 1 :]
    exponent = index - len(value) + 1
    return Decimal(integer.parse(integer_str, base), exponent, decimal_point=decimal_point)



from enum import IntEnum

from typing_extensions import Self

from lexorank.base import Base, Base36


class Sign(IntEnum):
    POSITIVE = 1
    NEGATIVE = -1

    def __neg__(self) -> Self:
        return self.__class__(-self.value)

    def __mul__(self, other: Self) -> Self:  # type: ignore[override]
        return self.__class__(self.value * other.value)


class Integer:
    def __init__(self, digits: list[int], sign: Sign = Sign.POSITIVE, base: Base = Base36) -> None:
        self._digits = self._lstrip(digits)
        self._sign = sign
        self._base = base

    @property
    def digits(self) -> list[int]:
        return self._digits.copy()

    @property
    def sign(self) -> Sign:
        return self._sign

    @property
    def base(self) -> Base:
        return self._base

    def __neg__(self) -> Self:
        return self.__class__(self._digits.copy(), -self._sign, self._base)

    def __abs__(self) -> Self:
        return self.__class__(self._digits.copy(), Sign.POSITIVE, self._base)

    def __add__(self, other: object) -> Self:
        other = self._type_guard(other)

        if self._sign != other._sign:
            return self - (-other)

        rdigits = []
        carry = 0

        for i in range(max(len(self), len(other))):
            a = self._digits[-i - 1] if i < len(self) else 0
            b = other._digits[-i - 1] if i < len(other) else 0

            result = a + b + carry
            carry = result // self._base.base()
            rdigits.append(result % self._base.base())

        if carry:
            rdigits.append(carry)

        return self.__class__(list(reversed(rdigits)), self._sign, self._base)

    def __sub__(self, other: object) -> Self:
        other = self._type_guard(other)

        if self._sign != other._sign:
            return self + (-other)

        a: Integer = abs(self)
        b = abs(other)
        sign = self._sign
        if a < b:
            a, b = b, a
            sign = -sign

        base = self._base.base()

        complement = [base - 1] * len(a)
        for i in range(len(b)):
            complement[-i - 1] = base - 1 - b._digits[-i - 1]

        rdigits = []
        carry = 1

        for i in range(len(a)):
            ad = a._digits[-i - 1]
            bd = complement[-i - 1]

            result = ad + bd + carry
            carry = result // base
            rdigits.append(result % base)

        return self.__class__(list(reversed(rdigits)), sign, self._base)

    def __mul__(self, other: object) -> Self:
        other = self._type_guard(other)

        rdigits = [0] * (len(self) + len(other))

        for i in range(len(self)):
            for j in range(len(other)):
                index = i + j
                ret = rdigits[index] + self._digits[-i - 1] * other._digits[-j - 1]
                rdigits[index] = ret % self._base.base()
                rdigits[index + 1] += ret // self._base.base()

        return self.__class__(list(reversed(rdigits)), self._sign * other._sign, self._base)

    def __lshift__(self, other: int) -> Self:
        if other < 0:
            raise ValueError("shift count must be non-negative")
        digits = self._digits + [0] * other
        return self.__class__(digits, self._sign, self._base)

    def __rshift__(self, other: int) -> Self:
        if other < 0:
            raise ValueError("shift count must be non-negative")
        return self.__class__(self._digits[: len(self) - other], self._sign, self._base)

    def __eq__(self, other: object):
        other = self._type_guard(other)

        if len(self) != len(other):
            return False
        if len(self) == 0:
            return True
        if self._sign != other._sign:
            return False
        for a, b in zip(self._digits, other._digits):
            if a != b:
                return False
        return True

    def __gt__(self, other: object):
        other = self._type_guard(other)

        if len(self) == 0 and len(other) == 0:
            return False
        if self._sign > other._sign:
            return True
        if self._sign < other._sign:
            return False
        if len(self) > len(other):
            return self._sign == Sign.POSITIVE
        if len(self) < len(other):
            return self._sign == Sign.NEGATIVE
        for a, b in zip(self._digits, other._digits):
            if a > b:
                return self._sign == Sign.POSITIVE
            if a < b:
                return self._sign == Sign.NEGATIVE
        return False

    def __ge__(self, other: object):
        return self.__gt__(other) or self.__eq__(other)

    def __lt__(self, other: object):
        return not self.__ge__(other)

    def __le__(self, other: object):
        return not self.__gt__(other)

    def __str__(self) -> str:
        if len(self) == 0:
            return "0"

        num = "".join([self._base.from_base10(digit) for digit in self._digits])
        return f"{'-' if self._sign == Sign.NEGATIVE else ''}{num}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} value={self}, base={self._base.base()}>"

    def __int__(self) -> int:
        value = 0
        for digit in self._digits:
            value = value * self._base.base() + digit
        return self._sign.value * value

    def __len__(self) -> int:
        return len(self._digits)

    def copy(self) -> Self:
        return self.__class__(self._digits.copy(), self._sign, self._base)

    @staticmethod
    def _lstrip(digits: list[int]) -> list[int]:
        for i, digit in enumerate(digits):
            if digit != 0:
                return digits[i:]

        return []

    def _type_guard(self, other: object) -> "Integer":
        if isinstance(other, (int, str)):
            return parse(other, self._base)
        if isinstance(other, Integer):
            return other
        raise ValueError(f"unsupported operand type(s) {type(other)}")


def parse(value: str | int, base: Base = Base36) -> Integer:
    if isinstance(value, int):
        sign = Sign.POSITIVE if value >= 0 else Sign.NEGATIVE
        value = abs(value)

        rdigits = []
        while value > 0:
            rdigits.append(value % base.base())
            value //= base.base()

        return Integer(list(reversed(rdigits)), sign, base)

    sign = Sign.POSITIVE if value[0] != "-" else Sign.NEGATIVE
    if value[0] in ("+", "-"):
        value = value[1:]

    digits = [base.to_base10(digit) for digit in value]
    return Integer(digits, sign, base)



from lexorank import lexorank
from lexorank.base import Base10, Base36, Base64
from lexorank.lexorank import Bucket


def test_parse():
    cases = [
        {"in": ("0|000012:3", Base10), "want": (Bucket.BUCEKT_0, "12:3", "0|000012:3")},
        {"in": ("0|00ri01:000000", Base36), "want": (Bucket.BUCEKT_0, "ri01:", "0|00ri01:")},
        {"in": ("1|000r^i:1", Base64), "want": (Bucket.BUCKET_1, "r^i:1", "1|000r^i:1")},
    ]

    for c in cases:
        got = lexorank.parse(*c["in"])
        assert got.bucket == c["want"][0]
        assert str(got.rank) == c["want"][1]
        assert str(got) == c["want"][2]


def test_middle():
    cases = [
        {"in": (Bucket.BUCEKT_0, Base10), "want": "0|500000:"},
        {"in": (Bucket.BUCEKT_0, Base36), "want": "0|i00000:"},
        {"in": (Bucket.BUCEKT_0, Base64), "want": "0|W00000:"},
    ]

    for c in cases:
        got = lexorank.middle(*c["in"])
        assert str(got) == c["want"]


def test_next():
    cases = [
        {"in": lexorank.middle(Bucket.BUCEKT_0, Base10), "want": "0|500016:"},
        {"in": lexorank.middle(Bucket.BUCEKT_0, Base36), "want": "0|i0000g:"},
        {"in": lexorank.middle(Bucket.BUCEKT_0, Base64), "want": "0|W0000G:"},
    ]

    for c in cases:
        got = c["in"].next()
        assert str(got) == c["want"]


def test_add():
    cases = [
        {"in": ("0|000001:", "0|000002:", Base36), "want": "0|000003:"},
    ]

    for c in cases:
        got = lexorank.parse(c["in"][0], c["in"][2])
        other = lexorank.parse(c["in"][1], c["in"][2])
        assert (got + other) == c["want"]


def test_eq():
    cases = [
        {"in": ("0|000001:", "0|000002:", Base36), "want": False},
        {"in": ("0|000001:", "0|000001:", Base36), "want": True},
        {"in": ("0|000002:", "0|000001:", Base36), "want": False},
    ]

    for c in cases:
        got = lexorank.parse(c["in"][0], c["in"][2])
        other = lexorank.parse(c["in"][1], c["in"][2])
        assert (got == other) == c["want"]


def test_gt():
    cases = [
        {"in": ("0|000001:", "0|000002:", Base36), "want": False},
        {"in": ("0|000001:", "0|000001:", Base36), "want": False},
        {"in": ("0|000002:", "0|000001:", Base36), "want": True},
        {"in": ("1|000001:", "0|000001:", Base36), "want": True},
    ]

    for c in cases:
        got = lexorank.parse(c["in"][0], c["in"][2])
        other = lexorank.parse(c["in"][1], c["in"][2])
        assert (got > other) == c["want"]



from enum import IntEnum
from functools import lru_cache

from typing_extensions import Self

from lexorank import decimal
from lexorank.base import Base, Base36
from lexorank.decimal import DECIMAL_POINT, Decimal
from lexorank.integer import Integer

BUCKET_SEPARATOR = "|"
WHOLE_NUMBER_SIZE = 6


class Bucket(IntEnum):
    BUCEKT_0 = 0
    BUCKET_1 = 1
    BUCKET_2 = 2

    def next(self) -> Self:
        return self.__class__((self.value + 1) % 3)


class LexoRank:
    def __init__(
        self,
        bucekt: Bucket,
        rank: Decimal,
        *,
        bucket_separator: str = BUCKET_SEPARATOR,
        whole_number_size: int = WHOLE_NUMBER_SIZE,
    ) -> None:
        self._bucket = bucekt
        self._rank = rank
        self._bucket_separator = bucket_separator
        self._whole_number_size = whole_number_size

    @property
    def bucket(self) -> Bucket:
        return self._bucket

    @property
    def rank(self) -> Decimal:
        return self._rank

    def prev(self, step: int = 16) -> Self:
        return self.next(-step)

    def next(self, step: int = 16) -> Self:
        return self.__class__(
            self._bucket,
            Decimal(self._rank.whole_number(), 0, decimal_point=self._rank.decimal_point) + step,
            bucket_separator=self._bucket_separator,
            whole_number_size=self._whole_number_size,
        )

    def __add__(self, other: object) -> Self:
        other = self._type_guard(other)

        return self.__class__(
            self._bucket,
            self._rank + other._rank,
            bucket_separator=self._bucket_separator,
            whole_number_size=self._whole_number_size,
        )

    def __sub__(self, other: object) -> Self:
        other = self._type_guard(other)

        return self.__class__(
            self._bucket,
            self._rank - other._rank,
            bucket_separator=self._bucket_separator,
            whole_number_size=self._whole_number_size,
        )

    def __mul__(self, other: object) -> Self:
        other = self._type_guard(other)

        return self.__class__(
            self._bucket,
            self._rank * other._rank,
            bucket_separator=self._bucket_separator,
            whole_number_size=self._whole_number_size,
        )

    def __eq__(self, other: object) -> bool:
        other = self._type_guard(other)

        return self._bucket == other._bucket and self._rank == other._rank

    def __gt__(self, other: object) -> bool:
        other = self._type_guard(other)

        if self._bucket.value > other._bucket.value:
            return True
        if self._bucket.value < other._bucket.value:
            return False
        return self._rank > other._rank

    def __ge__(self, other: object):
        return self.__gt__(other) or self.__eq__(other)

    def __lt__(self, other: object):
        return not self.__ge__(other)

    def __le__(self, other: object):
        return not self.__gt__(other)

    def __str__(self) -> str:
        rank = str(self._rank)
        index = rank.index(self._rank.decimal_point)

        whole = rank[:index].zfill(self._whole_number_size)
        rank = rank[index + 1 :]

        return (
            f"{self._bucket.value}{self._bucket_separator}{whole}{self._rank.decimal_point}{rank}"
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} value={self} base={self._rank.base.base()}>"

    def __float__(self) -> float:
        return float(self._rank)

    def __int__(self) -> int:
        return int(self._rank)

    def _type_guard(self, other: object) -> "LexoRank":
        if isinstance(other, LexoRank):
            return other
        if isinstance(other, str):
            return parse(
                other,
                self._rank.base,
                decimal_point=self._rank.decimal_point,
                bucket_separator=self._bucket_separator,
                whole_number_size=self._whole_number_size,
            )
        if isinstance(other, (int, float, Integer)):
            return LexoRank(self._bucket, decimal.parse(other, self._rank.base))
        if isinstance(other, Decimal):
            return LexoRank(self._bucket, other)
        raise ValueError(f"unsupported operand type(s) {type(other)}")


def parse(
    value: str,
    base: Base = Base36,
    *,
    decimal_point: str = DECIMAL_POINT,
    bucket_separator: str = BUCKET_SEPARATOR,
    whole_number_size: int = WHOLE_NUMBER_SIZE,
) -> LexoRank:
    if value[1] != bucket_separator or value[2 + whole_number_size] != decimal_point:
        raise ValueError(f"invalid lexorank format: {value}")

    bucket = Bucket(int(value[0]))
    decimal_ = decimal.parse(value[2:], base, decimal_point=decimal_point)
    return LexoRank(
        bucket, decimal_, bucket_separator=bucket_separator, whole_number_size=whole_number_size
    )


@lru_cache()
def middle(
    bucket: Bucket,
    base: Base = Base36,
    *,
    decimal_point: str = DECIMAL_POINT,
    bucket_separator: str = BUCKET_SEPARATOR,
    whole_number_size: int = WHOLE_NUMBER_SIZE,
) -> LexoRank:
    max_decimal = decimal.parse("1" + "0" * whole_number_size, base, decimal_point=decimal_point)
    return LexoRank(
        bucket,
        max_decimal * 0.5,
        bucket_separator=bucket_separator,
        whole_number_size=whole_number_size,
    )


def between(a: LexoRank | None, b: LexoRank | None) -> LexoRank:
    if a is None:
        if b is None:
            raise ValueError("a and b cannot be None at the same time")
        return b.prev()
    if b is None:
        return a.next()

    if a > b:
        a, b = b, a
    mid = (a + b) * 0.5
    if int(a) < int(mid) < int(b):
        return LexoRank(a.bucket, decimal.parse(mid.rank.whole_number()))

    # TODO: shorten the length of the decimal part
    return mid
